typedef unsigned char byte;
typedef unsigned char nibble; /* A nibble is 4 bits.
         One byte is two nibbles. =) */

/* The following assumes a 64-bit (8-byte) blocksize, with blocks
   represented as an array of 8 unsigned chars (bytes). Bit
   positions are indexed from 0-63. */

/* You should not need to directly use the next four macros */
#define BITMASK(b) (1 << (7-((b) % 8)))
#define BITSLOT(b) ((b) / 8)
#define BIT_ON(a,b) ((a)[BITSLOT(b)] |= BITMASK(b))
#define BIT_OFF(a,b) ((a)[BITSLOT(b)] &= (~BITMASK(b)))

// set the bth bit in block equal to c
#define BIT_SET(block,b,c) ((c) ? BIT_ON(block,b) : BIT_OFF(block,b))

// get the value of the bth bit in block
#define BIT_AT(block,b) (((block)[BITSLOT(b)] >> (7-((b)%8))) & 1)

/* Note: no input validation is done in cipher().
   It expects key to be a 16-byte string, and x to be
   an 8-byte string. As usual, buffer overflows are out of scope. */
void cipher(byte * key, byte * x);